"""Trace reader functions"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
import logging
import h5py
import numpy
import scipy.io
from neo import io

from . import igorpy
from .nwbreader import BBPNWBReader, ScalaNWBReader, AIBSNWBReader

logger = logging.getLogger(__name__)


def _check_metadata(metadata, reader_name, required_entries=[]):

    for entry in required_entries:

        if entry not in metadata:

            raise KeyError(
                "The trace reader {} expects the metadata {}. The "
                "entry {} was not provided.".format(
                    reader_name, ", ".join(e for e in required_entries), entry
                )
            )


def axon_reader(in_data):
    """Reader to read .abf
    Args:
        in_data (dict): of the format
        {
            "filepath": "./XXX.abf",
            "i_unit": "pA",
            "t_unit": "s",
            "v_unit": "mV",
        }
    """

    fp = in_data["filepath"]
    r = io.AxonIO(filename=fp)
    bl = r.read_block(lazy=False)

    data = []
    for trace in bl.segments:

        dt = 1.0 / int(trace.analogsignals[0].sampling_rate)
        np_trace = numpy.asarray(trace.analogsignals)

        if np_trace.shape[0] == 2:
            v = np_trace[0, :]
            c = np_trace[1, :]
        elif np_trace.shape[-1] == 2:
            v = np_trace[:, :, 0]
            c = np_trace[:, :, 1]
        else:
            raise Exception(f"Unknown .abf format for file {fp}. Maybe "
                            "it does not have current data?")

        data.append({
            "voltage": numpy.asarray(v).flatten(),
            "current": numpy.asarray(c).flatten(),
            "dt": dt
        })

    return data


def igor_reader(in_data):
    """Reader to read old .ibw

    Args:
        in_data (dict): of the format
        {
            'i_file': './XXX.ibw',
            'v_file': './XXX.ibw',
            'v_unit': 'V',
            't_unit': 's',
            'i_unit': 'A'
        }
    """

    _check_metadata(
        in_data, igor_reader.__name__, ["v_file", "i_file", "t_unit"]
    )

    # Read file
    notes_v, voltage = igorpy.read(in_data["v_file"])
    notes_i, current = igorpy.read(in_data["i_file"])

    if "A" in notes_v.dUnits and "V" in notes_i.dUnits:

        logger.warning(
            "It seems that the i_file and v_file are reversed for file: "
            "{}".format(in_data["v_file"])
        )

        voltage, current = current, voltage
        notes_v, notes_i = notes_i, notes_v

    # Extract data
    trace_data = {}
    trace_data["voltage"] = numpy.asarray(voltage)
    trace_data["v_unit"] = str(notes_v.dUnits).replace(" ", "")
    trace_data["dt"] = notes_v.dx
    trace_data["current"] = numpy.asarray(current)
    trace_data["i_unit"] = str(notes_i.dUnits).replace(" ", "")

    return [trace_data]


def read_matlab(in_data):
    """To read .mat from http://gigadb.org/dataset/100535

    Args:
        in_data (dict): of the format
        {
            'filepath': './161214_AL_113_CC.mat',
            'ton': 2000,
            'toff': 2500,
            'v_unit': 'V',
            't_unit': 's',
            'i_unit': 'A'
        }
    """

    _check_metadata(
        in_data,
        read_matlab.__name__,
        ["filepath", "i_unit", "v_unit", "t_unit"],
    )

    r = scipy.io.loadmat(in_data["filepath"])

    data = []
    for k, v in r.items():

        if "Trace" in k and k[-1] == "1":

            trace_data = {
                "current": v[:, 1],
                "voltage": r[k[:-1] + "2"][:, 1],
                "dt": v[1, 0],
            }

            data.append(trace_data)

    return data


def nwb_reader(in_data):
    """Reader for .nwb
    Args:
        in_data (dict): of the format
        {
            'filepath': './XXX.nwb',
            "protocol_name": "IV",
            "repetition": 1 (or [1, 3, ...]) # Optional
        }
    """

    _check_metadata(
        in_data,
        nwb_reader.__name__,
        ["filepath", "protocol_name"],
    )

    target_protocols = in_data['protocol_name']
    if isinstance(target_protocols, str):
        target_protocols = [target_protocols]

    with h5py.File(in_data["filepath"], "r") as content:
        if "data_organization" in content:
            reader = BBPNWBReader(content, target_protocols, in_data.get("repetition", None))
        elif "timeseries" in content["acquisition"].keys():
            reader = AIBSNWBReader(content, target_protocols)
        else:
            reader = ScalaNWBReader(content, target_protocols)

        data = reader.read()

    return data



from __future__ import annotations

"""Recording class"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
from abc import ABC, abstractmethod
import logging
import numpy
import efel
import matplotlib.pyplot as plt

from .tools import to_ms, to_mV, to_nA, set_efel_settings

logger = logging.getLogger(__name__)


class Recording(ABC):

    """Contains the data related to an electrophysiological recording."""

    def __init__(self, config_data, reader_data, protocol_name):
        """
        Constructor

        Args:
            config_data (dict): metadata for the recording considered informed
                by the user.
            reader_data (dict): metadata for the recording considered returned
                by the recording reader.
            protocol_name (str): name of the protocol of the present
                recording.
        """

        self.config_data = config_data
        self.reader_data = reader_data
        self.protocol_name = protocol_name

        if "filepath" in config_data:
            self.files = [config_data["filepath"]]
        elif "i_file" in config_data:
            self.files = [config_data["i_file"], config_data["v_file"]]
        else:
            self.files = []

        self.id = reader_data.get("id", None)

        self.location = None
        self.efeatures = {}

        self.t = None
        self.current = None
        self.voltage = None

        self.repetition = None

        if len(reader_data):
            self.t, self.current, self.voltage = self.standardize_trace(
                config_data, reader_data
            )

        self.export_attr = None
        self.auto_threshold = None
        self.peak_time = None

    @property
    def time(self):
        """Alias of the time attribute"""
        return self.t

    @time.setter
    def time(self, value):
        """Setter for an alias of the time attribute"""
        self.t = value

    @property
    def spikecount(self) -> int | None:
        if self.peak_time is None:
            return None
        else:
            return len(self.peak_time)

    def set_timing_ecode(self, name_timings, config_data):
        """Used by some of the children classes to check that the timing of
        the ecode is provided by the user and assign it to the correct
        attribute."""

        for timing in name_timings:
            if timing in config_data and config_data[timing] is not None:
                setattr(self, timing, int(round(config_data[timing] / self.dt)))
            else:
                raise AttributeError(
                    f"For protocol {self.protocol_name}, {timing} should"
                    "be specified in the config (in ms)."
                )

    def set_amplitudes_ecode(self, amp_name, config_data, reader_data, value):
        """Check in the user-provided data or reader-provided data if a
         given current amplitude is provided. If it isn't use the value
         computed from the current series"""

        if amp_name in config_data and config_data[amp_name] is not None:
            setattr(self, amp_name, config_data[amp_name])
        elif amp_name in reader_data and reader_data[amp_name] is not None:
            setattr(self, amp_name, reader_data[amp_name])
        else:
            setattr(self, amp_name, value)

    def index_to_ms(self, name_timing, time_series):
        """Used by some of the children classes to translate a timing attribute
         from index to ms."""

        setattr(self, name_timing, time_series[int(round(getattr(self, name_timing)))])

    def ms_to_index(self, timing):
        return int(round(timing / self.dt))

    def get_params(self):
        """Returns the eCode parameters"""
        return {attr: getattr(self, attr) for attr in self.export_attr}

    @abstractmethod
    def interpret(self):
        """Analyse a current array and extract from it the parameters needed to
        reconstruct the array"""
        pass

    @abstractmethod
    def generate(self):
        """Generate the current array from the parameters of the ecode"""
        pass

    @abstractmethod
    def compute_relative_amp(self, amp_threshold):
        """Divide all the amplitude in the stimuli by the spiking amplitude"""
        pass

    @abstractmethod
    def in_target(self, target, tolerance):
        """Returns a boolean. True if the amplitude of the eCode is close to
        target and False otherwise."""
        pass

    def standardize_trace(self, config_data, reader_data):
        """Standardize the units of the current and voltage times series. If
        some metadata are present both in the file itself and the file_metadata
         dictionary, the latter is used."""

        # Create the time series
        t = numpy.arange(len(reader_data["voltage"]))

        # Give it the correct sampling rate
        if "dt" in config_data and config_data["dt"] is not None:
            t = t * config_data["dt"]
        elif "dt" in reader_data and reader_data["dt"] is not None:
            t = t * reader_data["dt"]
        else:
            raise Exception(
                "Sampling rate not configured for "
                "file {}".format(self.files)
            )

        # Convert it to ms
        if "t_unit" in config_data and config_data["t_unit"] is not None:
            t = to_ms(t, config_data["t_unit"])
        elif "t_unit" in reader_data and reader_data["t_unit"] is not None:
            t = to_ms(t, reader_data["t_unit"])
        else:
            raise Exception(
                "Time unit not configured for " "file {}".format(self.files)
            )

        # Convert current to nA
        if "i_unit" in config_data and config_data["i_unit"] is not None:
            current = to_nA(reader_data["current"], config_data["i_unit"])
        elif "i_unit" in reader_data and reader_data["i_unit"] is not None:
            current = to_nA(reader_data["current"], reader_data["i_unit"])
        else:
            raise Exception(
                "Current unit not configured for " "file {}".format(self.files)
            )

        # Convert voltage to mV
        if "v_unit" in config_data and config_data["v_unit"] is not None:
            voltage = to_mV(reader_data["voltage"], config_data["v_unit"])
        elif "v_unit" in reader_data and reader_data["v_unit"] is not None:
            voltage = to_mV(reader_data["voltage"], reader_data["v_unit"])
        else:
            raise Exception(
                "Voltage unit not configured for " "file {}".format(self.files)
            )

        # Offset membrane potential to known value
        if "v_corr" in config_data and config_data["v_corr"] is not None:
            voltage = (
                voltage - numpy.median(voltage[:100]) + config_data["v_corr"]
            )

        # Correct for the liquid junction potential
        # WARNING: the ljp is informed as a positive float but we substract it
        # from the voltage
        if "ljp" in config_data and config_data["ljp"] is not None:
            voltage = voltage - config_data["ljp"]

        if "repetition" in reader_data:
            self.repetition = reader_data["repetition"]

        return t, current, voltage

    def call_efel(self, efeatures, efel_settings=None):
        """ Calls efel to compute the wanted efeatures """

        if efel_settings is None:
            efel_settings = {}

        settings = {"stimulus_current": self.amp}

        if "Threshold" not in efel_settings and self.auto_threshold is not None:
            logger.warning(f"Threshold was not provided and was automatically"
                           f" set to {self.auto_threshold}")
            settings["Threshold"] = self.auto_threshold

        for setting in efel_settings:
            if setting not in ['stim_start', 'stim_end']:
                settings[setting] = efel_settings[setting]

        stim_start = efel_settings.get('stim_start', self.ton)
        stim_end = efel_settings.get('stim_end', self.toff)

        # Special case for SpikeRec
        if "multiple_decay_time_constant_after_stim" in efeatures:
            if hasattr(self, "multi_stim_start") and hasattr(self, "multi_stim_end"):
                settings["multi_stim_start"] = self.multi_stim_start
                settings["multi_stim_end"] = self.multi_stim_end
            elif "stim_start" in efel_settings and "stim_end" in efel_settings:
                settings["multi_stim_start"] = [stim_start]
                settings["multi_stim_end"] = [stim_end]

        set_efel_settings(settings)

        efel_trace = {
            "T": self.t,
            "V": self.voltage,
            'stim_start': [stim_start],
            'stim_end': [stim_end]
        }

        try:
            return efel.getFeatureValues(
                [efel_trace], efeatures, raise_warnings=False
            )
        except TypeError as e:
            if "Unknown feature name" in str(e):
                str_f = " ".join(efeatures)
                raise Exception("One of the following feature name does not "
                                f"exist in eFEL: {str_f}")

    def compute_efeatures(
        self, efeatures, efeature_names=None, efel_settings=None
    ):
        """Compute a set of efeatures for the present recording.

        Args:
            efeatures (list of str): name of the efeatures to extract from
                the recordings.
            efeature_names (list of str): Optional. Given name for the
                features. Can and should be used if the same feature
                is to be extracted several time on different sections
                of the same recording.
            efel_settings (dict): eFEL settings in the form
                {setting_name: setting_value}.
        """
        if efeature_names is None:
            efeature_names = efeatures

        for i, f in enumerate(efeatures):
            if efeature_names[i] is None:
                efeature_names[i] = f

        efel_vals = self.call_efel(efeatures, efel_settings)
        for efeature_name, efeature in zip(efeature_names, efeatures):

            value = efel_vals[0][efeature]
            if value is None or len(value) == 0 or numpy.isinf(numpy.nanmean(value)):
                self.efeatures[efeature_name] = numpy.nan
            else:
                self.efeatures[efeature_name] = numpy.nanmean(value)

    def compute_spikecount(self, efel_settings=None):
        """Compute the number of spikes in the trace"""
        if not efel_settings:
            efel_settings = {}

        tmp_settings = {'strict_stiminterval': True}
        if efel_settings is not None:
            tmp_settings.update(efel_settings)

        # If the setting Threshold is not provided, tries to find it
        if tmp_settings.get("Threshold", None) is None:
            tmp_settings["Threshold"] = self.auto_threshold
            efel_vals = self.call_efel(['peak_time'], tmp_settings)
            self.peak_time = efel_vals[0]['peak_time']

        else:
            self.peak_time = self.call_efel(['peak_time'], tmp_settings)[0]['peak_time']

        if self.spikecount == 0 and numpy.max(self.voltage) > tmp_settings["Threshold"]:
            logger.warning(
                f"No spikes were detected in recording {self.files} but the "
                "voltage goes higher than the spike detection threshold."
            )

    def set_autothreshold(self, offset_voltage=20.) -> None:
        """Computes the threshold based on the input voltage sets it as an attribute."""
        idx_ton = self.ms_to_index(self.ton)
        idx_toff = self.ms_to_index(self.toff)
        step_voltage = numpy.median(self.voltage[idx_ton:idx_toff])
        base_voltage = numpy.median(self.voltage[:idx_ton])
        if base_voltage > step_voltage:
            thresh = base_voltage + offset_voltage
        else:
            thresh = step_voltage + offset_voltage
        # The threshold cannot be lower than the base voltage (handles the case
        # where the step is hyperpolarizing)
        self.auto_threshold = numpy.clip(thresh, base_voltage + offset_voltage, 50.)

    def in_target(self, target, tolerance, absolute_amplitude=False):
        """Returns a boolean. True if the amplitude of the eCode is close to
        target and False otherwise."""

        effective_amp = self.amp if absolute_amplitude else self.amp_rel

        if numpy.abs(target - effective_amp) < tolerance:
            return True

        return False

    def compute_relative_amp(self, amp_threshold):
        """Divide all the amplitude in the stimuli by the spiking amplitude"""

        self.amp_rel = 100.0 * self.amp / amp_threshold
        self.hypamp_rel = 100.0 * self.hypamp / amp_threshold

    def get_plot_amplitude_title(self):
        return " ({:.01f}%)".format(self.amp_rel)

    def plot(
        self,
        axis_current=None,
        axis_voltage=None,
        display_xlabel=True,
        display_ylabel=True
    ):
        """Plot the recording"""

        if axis_current is None or axis_voltage is None:
            _, axs = plt.subplots(nrows=2, ncols=1, figsize=[4.9, 4.8])
            axis_current, axis_voltage = axs[0], axs[1]

        title = "Holding Amp = {:.03f} nA\nAmp = {:.03f} nA".format(
            self.hypamp, self.amp)
        if self.amp_rel is not None:
            title += self.get_plot_amplitude_title()
        if self.id is not None:
            title += "\nid: {}".format(self.id)
        if self.repetition is not None:
            title += "\nRepetition: {}".format(self.repetition)
        axis_current.set_title(title, size="x-small")

        gen_t, gen_i = self.generate()
        axis_current.plot(self.t, self.current, c="C0", lw=0.8)
        axis_current.plot(gen_t, gen_i, c="C1", ls="--", lw=0.8)
        axis_voltage.plot(self.t, self.voltage, c="C0", lw=0.8)

        if self.peak_time is not None:
            max_v = numpy.max(self.voltage)
            for pt in self.peak_time:
                axis_voltage.plot(
                    [pt, pt], [max_v + 5, max_v + 15], c="C3", ls="-", lw=0.5
                )

        if self.auto_threshold is not None:
            axis_voltage.plot(
                [self.t[0], self.t[-1]], [self.auto_threshold, self.auto_threshold],
                c="black", ls="--", lw=0.5, alpha=0.8
            )

        if display_xlabel:
            axis_voltage.set_xlabel("Time (ms)")
        if display_ylabel:
            axis_current.set_ylabel("Current (nA)")
            axis_voltage.set_ylabel("Voltage (mV)")

        axis_current.tick_params(axis="both", which="major", labelsize=8)
        axis_current.tick_params(axis="both", which="minor", labelsize=6)
        axis_voltage.tick_params(axis="both", which="major", labelsize=8)
        axis_voltage.tick_params(axis="both", which="minor", labelsize=6)

        return axis_current, axis_voltage



"""Efeature extraction functions"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
import os
import pickle
import functools
import logging
import pathlib
import gc

import numpy

from bluepyefe import tools
from bluepyefe.cell import Cell
from bluepyefe.plotting import plot_all_recordings
from bluepyefe.plotting import plot_all_recordings_efeatures
from bluepyefe.protocol import Protocol
from bluepyefe.target import EFeatureTarget
from bluepyefe.rheobase import compute_rheobase_absolute
from bluepyefe.rheobase import compute_rheobase_flush
from bluepyefe.rheobase import compute_rheobase_majority_bin
from bluepyefe.rheobase import compute_rheobase_interpolation
from bluepyefe.tools import DEFAULT_EFEL_SETTINGS, PRESET_PROTOCOLS_RHEOBASE
from bluepyefe.auto_targets import default_auto_targets

logger = logging.getLogger(__name__)


def _create_cell(cell_definition, recording_reader, efel_settings=None):
    """
    Initialize a Cell object and populate it with the content of the associated
    recording files.

    The present function exist to be use by the map_function.

    Args:
        cell_definition (dict): subpart of files_metadata containing the
            information for a single cell.
        recording_reader (callable or None): method that will be used to
            read the files containing the recordings. If None, the function
            used will be chosen automatically based on the extension
            of the file.
        efel_settings (dict): eFEL settings in the form
            {setting_name: setting_value}.
    """

    cell_name = cell_definition[0]
    cell = cell_definition[1]

    out_cell = Cell(name=cell_name)

    for prot_name in cell.keys():
        out_cell.read_recordings(
            protocol_data=cell[prot_name],
            protocol_name=prot_name,
            recording_reader=recording_reader,
            efel_settings=efel_settings
        )

    return out_cell


def _extract_efeatures_cell(cell, targets, efel_settings=None):
    """
    Compute the efeatures on all the recordings of a Cell.

    The present function exists to be use by the map_function.

    Args:
        cell (Cell): cell for which to extract the efeatures.
        targets (list of Target): targets to extract from the recordings of
            the present cell.
        efel_settings (dict): eFEL settings in the form
            {setting_name: setting_value}.
    """

    if efel_settings is None:
        efel_settings = {}

    # Group targets per same protocol and same eFEL settings for efficiency
    setting_groups = []
    for target in targets:

        efeature_name = target.get("efeature_name", target["efeature"])

        for i, group in enumerate(setting_groups):

            if target["efel_settings"] == group['efel_settings'] and \
                    target["protocol"] == group['protocol']:
                setting_groups[i]["efeatures"].append(target["efeature"])
                setting_groups[i]['efeature_names'].append(efeature_name)
                break

        else:

            setting_group = {
                'efel_settings': target["efel_settings"],
                'protocol': target["protocol"],
                'efeatures': [target["efeature"]],
                'efeature_names': [efeature_name]
            }
            setting_groups.append(setting_group)

    for group in setting_groups:
        cell.extract_efeatures(
            group['protocol'],
            group["efeatures"],
            group["efeature_names"],
            efel_settings={**efel_settings, **group["efel_settings"]}
        )

    return cell


def _saving_data(output_directory, feat, stim, currents):
    """
    Save the features, protocols and current to json files.

    Args:
        output_directory (str): path of the directory to which the efeatures
            and protocols data will be saved.
        feat (dict): contains the information related to the efeatures.
        stim (dict): contains the information related to the protocols.
        currents (dict): contains the information related to the holding
            and threshold currents.
    """

    output_directory = pathlib.Path(output_directory)
    output_directory.mkdir(parents=True, exist_ok=True)

    features_path = output_directory / "features.json"
    tools.dict_to_json(feat, features_path)
    logger.info("Saved efeatures to {}".format(features_path))

    protocols_path = output_directory / "protocols.json"
    tools.dict_to_json(stim, protocols_path)
    logger.info("Saving protocols to {}".format(protocols_path))

    currents_path = output_directory / "holding_threshold_currents.json"
    tools.dict_to_json(currents, currents_path)
    logger.info(
        "Saving threshold and holding currents to {}" "".format(currents_path)
    )


def read_recordings(
    files_metadata,
    recording_reader=None,
    map_function=map,
    efel_settings=None
):
    """
    Read recordings from a group of files. The files are expected to be
    identified by both a cell id and a protocol name (see files_metadata).

    Args:
        files_metadata (dict): define for which cell and protocol each file
            has to be used. Of the form:
            {
                cell_id: {
                    protocol_name: [{file_metadata1}, {file_metadata1}]
                }
            }
            A same file path might be present in the file metadata for
            different protocols.
            The entries required in the file_metadata are specific to each
            recording_reader (see bluepyemodel/reader.py to know which one are
            needed for your recording_reader).
        recording_reader (function): custom recording reader function.
            It's inner working has to match the metadata entered in
            files_metadata.
        map_function (function): Function used to map (parallelize) the
            recording reading operations. Note: the parallelization is
            done across cells an not across files.
        efel_settings (dict): eFEL settings in the form
            {setting_name: setting_value}.

    Return:
         cells (list): list of Cell objects containing the data of the
         recordings
    """

    cells = map_function(
        functools.partial(
            _create_cell,
            recording_reader=recording_reader,
            efel_settings=efel_settings
        ),
        list(files_metadata.items()),
    )

    return list(cells)


def extract_efeatures_at_targets(
    cells,
    targets,
    map_function=map,
    efel_settings=None
):
    """
    Extract efeatures from recordings following the protocols, amplitudes and
    efeature names specified in the targets.

    Args:
        cells (list): list of Cells containing the recordings from which the
            efeatures will be extracted.
        targets (dict): define the efeatures to extract as well as which
            protocols and current amplitude they should be extracted for. Of
            the form:
            [{
                "efeature": "AP_amplitude",
                "protocol": "IDRest",
                "amplitude": 150.,
                "tolerance": 10.,
                "efel_settings": {
                    'stim_start': 200.,
                    'stim_end': 500.,
                    'Threshold': -10.
                }
            }]
        map_function (function): Function used to map (parallelize) the
            feature extraction operations. Note: the parallelization is
            done across cells an not across efeatures.
        efel_settings (dict): eFEL settings in the form
            {setting_name: setting_value}. If settings are also informed
            in the targets per efeature, the latter will have priority.
    """

    for target in targets:

        if 'location' not in target:
            target["location"] = "soma"

        if 'efel_settings' not in target:
            target['efel_settings'] = {}

    cells = map_function(
        functools.partial(
            _extract_efeatures_cell,
            targets=targets,
            efel_settings=efel_settings
        ),
        cells,
    )

    return list(cells)


def compute_rheobase(
    cells,
    protocols_rheobase,
    rheobase_strategy="absolute",
    rheobase_settings=None
):
    """
    For each cell, finds the smallest current inducing a spike (rheobase).
    This currents are then use it to compute the relative amplitude of
    the stimuli.

    Args:
        cells (list): list of Cells containing for which the rheobase will be
            computed
        protocols_rheobase (list): names of the protocols that will be
            used to compute the rheobase of the cells. E.g: ['IDthresh'].
        rheobase_strategy (str): function used to compute the rheobase. Can be
            'absolute' (amplitude of the lowest amplitude inducing at least a
            spike) or 'majority' (amplitude of the bin in which a majority of
            sweeps induced at least one spike).
        rheobase_settings (dict): settings related to the rheobase computation.
            Keys have to match the arguments expected by the rheobase
            computation function.
    """

    if rheobase_settings is None:
        rheobase_settings = {}

    if rheobase_strategy == "absolute":
        rheobase_function = compute_rheobase_absolute
    elif rheobase_strategy == "flush":
        rheobase_function = compute_rheobase_flush
    elif rheobase_strategy == "majority":
        rheobase_function = compute_rheobase_majority_bin
    elif rheobase_strategy == "interpolation":
        rheobase_function = compute_rheobase_interpolation
    else:
        raise Exception(f"Rheobase strategy {rheobase_strategy} unknown.")

    for cell in cells:
        rheobase_function(cell, protocols_rheobase, **rheobase_settings)
        cell.compute_relative_amp()


def _build_protocols(
    targets,
    global_rheobase,
    protocol_mode,
    efel_settings=None
):
    """Build a list of Protocols that matches the expected targets"""

    if efel_settings is None:
        efel_settings = {}

    protocols = []

    for target in targets:

        settings = {**efel_settings, **target.get('efel_settings', {})}
        efeature_name = target.get("efeature_name", target["efeature"])

        efeature_target = EFeatureTarget(
            efeature_name=efeature_name,
            efel_feature_name=target['efeature'],
            protocol_name=target['protocol'],
            amplitude=target['amplitude'],
            tolerance=target['tolerance'],
            efel_settings=settings,
        )

        for i, p in enumerate(protocols):
            if p.name == target['protocol'] and \
                    p.amplitude == target['amplitude']:
                protocols[i].feature_targets.append(efeature_target)
                break
        else:
            protocols.append(
                Protocol(
                    name=target['protocol'],
                    amplitude=target['amplitude'],
                    tolerance=target['tolerance'],
                    feature_targets=[efeature_target],
                    global_rheobase=global_rheobase,
                    mode=protocol_mode
                )
            )

    return protocols


def group_efeatures(
    cells,
    targets,
    absolute_amplitude=False,
    use_global_rheobase=True,
    protocol_mode='mean',
    efel_settings=None
):
    """
    Group the recordings and their efeatures and associate them to the
    EFeature Targets and Protocols they belong to.

    Args:
        cells (list): list of Cells containing for which the rheobase will be
            computed
        targets (dict): define the efeatures to extract as well as which
            protocols and current amplitude they should be extracted for. Of
            the form:
            [{
                "efeature": "AP_amplitude",
                "protocol": "IDRest",
                "amplitude": 150.,
                "tolerance": 10.,
                "efel_settings": {
                    'stim_start': 200.,
                    'stim_end': 500.,
                    'Threshold': -10.
                }
            }]
        absolute_amplitude (bool): if True, will use the absolute amplitude
            instead of the relative amplitudes of the recordings when checking
            if a recording has to be used for a given target.
        use_global_rheobase (bool): As the final amplitude of a target is the
            mean of the amplitude of the cells, a global rheobase can be used
            to avoid issues when a cell matches a target but not another one.
            Which can result in situations where the second target is higher
            but it's amp in amperes is lower.
            e.g: target1 = 100%, target2 = 150% but target1_amp = 0.2 and
            target2_amp = 0.18
        protocol_mode (mean): if a protocol matches several recordings, the
            mode set the logic of how the output will be generating. Must be
            'mean', 'median' or 'lnmc'
        efel_settings (dict): eFEL settings in the form
            {setting_name: setting_value}. If settings are also informed
            in the targets per efeature, the latter will have priority.
    """

    if efel_settings is None:
        efel_settings = {}

    global_rheobase = None
    if use_global_rheobase and not absolute_amplitude:
        global_rheobase = numpy.nanmean(
            [c.rheobase for c in cells if c.rheobase is not None]
        )

    protocols = _build_protocols(
        targets,
        global_rheobase=global_rheobase,
        protocol_mode=protocol_mode,
        efel_settings=efel_settings,
    )

    for protocol in protocols:
        for cell in cells:

            if cell.rheobase is None and not absolute_amplitude:
                continue

            for recording in cell.get_recordings_by_protocol_name(
                    protocol.name
            ):

                if recording.in_target(
                    protocol.amplitude,
                    protocol.tolerance,
                    absolute_amplitude
                ):
                    protocol.append(recording)

    return protocols


def _build_current_dict(cells, default_std_value):
    """
    Compute the mean and standard deviation of the holding and threshold
    currents.
    """

    holding = {}
    threshold = {}

    for cell in cells:

        holding[cell.name] = numpy.nanmean(
            [t.hypamp for t in cell.recordings]
        )

        if cell.rheobase is not None:
            threshold[cell.name] = cell.rheobase

    std_holding = numpy.nanstd(list(holding.values()))
    if std_holding == 0:
        std_holding = default_std_value

    std_threshold = numpy.nanstd(list(threshold.values()))
    if std_threshold == 0:
        std_threshold = default_std_value

    currents = {
        "holding_current": [
            numpy.nanmean(list(holding.values())),
            std_holding,
        ],
        "threshold_current": [
            numpy.nanmean(list(threshold.values())),
            std_threshold,
        ],
        "all_holding_current": holding,
        "all_threshold_current": threshold,
    }

    return currents


def create_feature_protocol_files(
    cells,
    protocols,
    output_directory=None,
    threshold_nvalue_save=1,
    write_files=True,
    save_files_used=False,
    default_std_value=1e-3
):
    """
    Save the efeatures and protocols for each protocol/target combo
    in json file.

    Args:
        cells (list): list of Cells.
        protocols (list): list of Protocols.
        output_directory (str): path of the directory to which the features,
            protocols and currents will be saved.
        threshold_nvalue_save (int): minimum number of values needed for
            an efeatures to be averaged and returned in the output.
        write_files (bool): if True, the efeatures, protocols and currents
            will be saved in .json files in addition of being returned.
        save_files_used (bool): if True, the name of the recording files used
            in the computation of the features will be added to the
            efeatures.
        default_std_value (float): default value used to replace the standard
            deviation if the standard deviation is 0.

    Returns:
        feat (dict)
        stim (dict)
        currents (dict)

    """

    out_features = {}
    out_stimuli = {}

    for protocol in protocols:

        stimname = protocol.stimulus_name

        tmp_feat = []

        for target in protocol.feature_targets:

            if target.sample_size < threshold_nvalue_save:
                logger.warning(
                    "Number of values < threshold_nvalue_save for efeature "
                    "{} stimulus {}. The efeature will be ignored"
                    "".format(target.efel_feature_name, stimname)
                )
                continue

            tmp_feat.append(target.as_dict(save_files_used, default_std_value))

        if not tmp_feat:
            logger.warning(
                "No efeatures for stimulus {}. The protocol will not "
                "be created.".format(stimname)
            )
            continue

        out_features[stimname] = {'soma': tmp_feat}
        out_stimuli[stimname] = protocol.as_dict()

    # Compute the mean and std of holding and threshold currents
    currents = _build_current_dict(cells, default_std_value)

    if write_files:

        if not output_directory:
            raise Exception(
                f"output_directory cannot be {output_directory}"
                f" if write_files is True."
            )

        _saving_data(
            output_directory,
            out_features,
            out_stimuli,
            currents
        )

    return out_features, out_stimuli, currents


def _read_extract(
    files_metadata,
    recording_reader,
    map_function,
    targets,
    efel_settings=None
):
    """Read recordings and create the matching Cell objects based on a files_metadata."""

    cells = read_recordings(
        files_metadata,
        recording_reader=recording_reader,
        map_function=map_function,
        efel_settings=efel_settings
    )

    cells = extract_efeatures_at_targets(
        cells, targets, map_function=map_function, efel_settings=efel_settings
    )

    return cells


def _read_extract_low_memory(
    files_metadata, recording_reader, targets, efel_settings=None
):
    """Read recordings and create the matching Cell objects based on a
    files_metadata. Does not us a map function and delete the recording's
    data on the go to avoid using too much memory."""

    cells = []
    for cell_name in files_metadata:

        cell = read_recordings(
            {cell_name: files_metadata[cell_name]},
            recording_reader=recording_reader,
            efel_settings=efel_settings
        )[0]

        extract_efeatures_at_targets(
            [cell], targets, map, efel_settings
        )

        # clean traces voltage and time
        for i in range(len(cell.recordings)):
            cell.recordings[i].t = None
            cell.recordings[i].voltage = None
            cell.recordings[i].current = None
            cell.recordings[i].reader_data = None

        cells.append(cell)
        gc.collect()

    return cells


def convert_legacy_targets(targets):
    """Convert targets of the form:
        protocol_name: {
                "amplitudes": [50, 100],
                "tolerances": [10, 10],
                "efeatures": {"Spikecount": {'Threshold': -10.}},
                "location": "soma"
            }
        }
    To ones of the form:
        [{
            "efeature": "AP_amplitude",
            "protocol": "IDRest",
            "amplitude": 150.,
            "tolerance": 10.,
            "efel_settings": {
                'stim_start': 200.,
                'stim_end': 500.,
                'Threshold': -10.
            }
        }]
    """

    formatted_targets = []

    for protocol_name, target in targets.items():
        for i, amplitude in enumerate(target["amplitudes"]):

            tolerances = target["tolerances"]
            if len(tolerances) == 1:
                tolerances = tolerances * len(target["amplitudes"])

            for efeature in target["efeatures"]:

                formatted_target = {
                    "efeature": efeature,
                    "protocol": protocol_name,
                    "amplitude": amplitude,
                    "tolerance": tolerances[i],
                    "efel_settings": {}
                }

                if isinstance(target["efeatures"], dict):
                    formatted_target["efel_settings"] = target[
                        "efeatures"][efeature]

                formatted_targets.append(formatted_target)

    return formatted_targets


def _extract_with_targets(
    files_metadata,
    targets=None,
    protocols_rheobase=None,
    absolute_amplitude=False,
    recording_reader=None,
    map_function=map,
    low_memory_mode=False,
    protocol_mode="mean",
    efel_settings=None,
    rheobase_strategy="absolute",
    rheobase_settings=None
):
    """Read the recordings and extract the efeatures at the requested
    targets"""

    if not low_memory_mode:
        cells = _read_extract(
            files_metadata,
            recording_reader,
            map_function,
            targets,
            efel_settings
        )
    else:
        cells = _read_extract_low_memory(
            files_metadata, recording_reader, targets, efel_settings
        )

    if not absolute_amplitude:
        compute_rheobase(
            cells,
            protocols_rheobase=protocols_rheobase,
            rheobase_strategy=rheobase_strategy,
            rheobase_settings=rheobase_settings
        )

    protocols = group_efeatures(
        cells,
        targets,
        absolute_amplitude=absolute_amplitude,
        use_global_rheobase=True,
        protocol_mode=protocol_mode,
        efel_settings=efel_settings
    )

    return cells, protocols


def _extract_auto_targets(
    files_metadata,
    protocols_rheobase,
    recording_reader,
    map_function,
    protocol_mode,
    efel_settings,
    auto_targets=None,
    rheobase_strategy="flush",
    rheobase_settings=None
):
    """Read the recordings and extract the efeatures using AutoTargets"""

    if rheobase_settings is None and rheobase_strategy == "flush":
        rheobase_settings = {"upper_bound_spikecount": 4}

    if auto_targets is None:
        auto_targets = default_auto_targets()

    cells = read_recordings(
        files_metadata,
        recording_reader=recording_reader,
        map_function=map_function,
        efel_settings=efel_settings
    )

    compute_rheobase(
        cells,
        protocols_rheobase=protocols_rheobase,
        rheobase_strategy=rheobase_strategy,
        rheobase_settings=rheobase_settings
    )

    if not sum(bool(c.rheobase) for c in cells):
        logger.warning("No cells with valid rheobase")
        return cells, [], []

    recordings = []
    for c in cells:
        recordings += c.recordings

    for i in range(len(auto_targets)):
        auto_targets[i].select_ecode_and_amplitude(recordings)

    # Extract the efeatures and group them around the preset of targets.
    targets = []
    for at in auto_targets:
        targets += at.generate_targets()

    cells = extract_efeatures_at_targets(
        cells, targets, map_function=map_function, efel_settings=efel_settings
    )

    protocols = group_efeatures(
        cells,
        targets,
        use_global_rheobase=True,
        protocol_mode=protocol_mode,
        efel_settings=efel_settings
    )

    return cells, protocols, targets


def extract_efeatures_per_cell(
    files_metadata,
    cells,
    output_directory,
    targets,
    protocol_mode,
    threshold_nvalue_save,
    write_files,
    default_std_value=1e-3
):

    for cell_name in files_metadata:

        cell_directory = str(pathlib.Path(output_directory) / cell_name)

        for cell in cells:

            if cell.name != cell_name:
                continue

            cell_protocols = group_efeatures(
                [cell],
                targets,
                use_global_rheobase=True,
                protocol_mode=protocol_mode
            )

            _ = create_feature_protocol_files(
                [cell],
                cell_protocols,
                output_directory=cell_directory,
                threshold_nvalue_save=threshold_nvalue_save,
                write_files=write_files,
                default_std_value=default_std_value
            )


def extract_efeatures(
    output_directory,
    files_metadata,
    targets=None,
    threshold_nvalue_save=1,
    protocols_rheobase=None,
    absolute_amplitude=False,
    recording_reader=None,
    map_function=map,
    write_files=False,
    plot=False,
    low_memory_mode=False,
    protocol_mode="mean",
    efel_settings=None,
    extract_per_cell=False,
    rheobase_strategy="absolute",
    rheobase_settings=None,
    auto_targets=None,
    pickle_cells=False,
    default_std_value=1e-3
):
    """
    Extract efeatures.

    Args:
        output_directory (str): path to the output directory
        files_metadata (dict): define from files to read the data as well as
            the name of the cells and protocols to which these data are
            related. Of the form:
            {
                cell_id: {
                    protocol_name: [{file_metadata1}, {file_metadata1}]
                }
            }
            The entries required in the file_metadata are specific to each
            recording_reader (see bluepyemodel/reader.py to know which one are
            needed for your recording_reader).
            As the file_metadata contain file paths, a same file path might
            need to be present in the file metadata of different protocols if
            the path contains data coming from different stimuli (eg: for NWB).
        targets (list): define the efeatures to extract as well as which
            protocols and current amplitude (expressed either in % of the
            rheobase if absolute_amplitude if False or in nA if
            absolute_amplitude is True) they should be extracted.
            If targets are not provided, automatic targets will be used.
            Of the form:
            [{
                "efeature": "AP_amplitude",
                "protocol": "IDRest",
                "amplitude": 150.,
                "tolerance": 10.,
                "efel_settings": {
                    'stim_start': 200.,
                    'stim_end': 500.,
                    'Threshold': -10.
                }
            }]
        threshold_nvalue_save (int): minimum number of values needed for
            an efeatures to be averaged and returned in the output.
        protocols_rheobase (list): names of the protocols that will be
            used to compute the rheobase of the cells. E.g: ['IDthresh'].
        absolute_amplitude (bool): if True, will use the absolute amplitude
            instead of the relative amplitudes of the recordings when checking
            if a recording has to be used for a given target.
        recording_reader (function): custom recording reader function. It's
            inner working has to match the metadata entered in files_metadata.
        map_function (function): Function used to map (parallelize) the
            recording reading and feature extraction operations.
        write_files (bool): if True, the efeatures, protocols and currents
            will be saved in .json files in addition of being returned.
        plot (bool): if True, the recordings and efeatures plots will be
            created.
        low_memory_mode (bool): if True, minimizes the amount of memory used
            during the data reading and feature extraction steps by performing
            additional clean up. Not compatible with map_function.
        protocol_mode (str): protocol_mode (mean): if a protocol matches
            several recordings, the mode set the logic of how the output
            will be generating. Must be 'mean', 'median' or 'lnmc'
        efel_settings (dict): eFEL settings in the form
            {setting_name: setting_value}. If settings are also informed
            in the targets per efeature, the latter will have priority. If
            None, will be set to:
            {
                'strict_stiminterval': True,
                'Threshold': -20.,
                'interp_step': 0.025
            }
        extract_per_cell (bool): if True, also generates the features.json and
            protocol.json for each individual cells.
        rheobase_strategy (str): function used to compute the rheobase. Can be
            'absolute' (amplitude of the lowest amplitude inducing at least a
            spike) or 'majority' (amplitude of the bin in which a majority of
            sweeps induced at least one spike).
        rheobase_settings (dict): settings related to the rheobase computation.
            Keys have to match the arguments expected by the rheobase
            computation function.
        auto_targets (list of AutoTarget): targets with more flexible goals.
        pickle_cells (bool): if True, the cells object will be saved as a pickle file.
        default_std_value (float): default value used to replace the standard
            deviation if the standard deviation is 0.
    """

    if not files_metadata:
        raise ValueError("Argument 'files_metadata' is empty")

    if efel_settings is None:
        logger.warning(
            "efel_settings is None. Default settings will be used"
        )
        efel_settings = DEFAULT_EFEL_SETTINGS.copy()

    if protocols_rheobase is None and not absolute_amplitude:
        logger.warning(
            "protocols_rheobase is None. Default protocol names will be used"
        )
        protocols_rheobase = PRESET_PROTOCOLS_RHEOBASE.copy()

    if low_memory_mode and map_function not in [map, None]:
        logger.warning(
            "low_memory_mode is not compatible with the use of map_function"
        )
    if low_memory_mode and plot:
        raise Exception('plot cannot be used in low_memory_mode mode.')

    if targets is not None and isinstance(targets, dict):
        logger.warning(
            "targets seems to be in a legacy format. A conversion will"
            " be performed."
        )
        targets = convert_legacy_targets(targets)

    if targets is not None and auto_targets is not None:
        raise Exception("Cannot specify both targets and auto_targets.")

    if (
        not absolute_amplitude and
        (targets is None or auto_targets is not None)
    ):
        cells, protocols, targets = _extract_auto_targets(
            files_metadata,
            protocols_rheobase,
            recording_reader,
            map_function,
            protocol_mode,
            efel_settings,
            auto_targets,
            rheobase_strategy,
            rheobase_settings
        )
    else:
        cells, protocols = _extract_with_targets(
            files_metadata,
            targets,
            protocols_rheobase,
            absolute_amplitude,
            recording_reader,
            map_function,
            low_memory_mode,
            protocol_mode,
            efel_settings,
            rheobase_strategy,
            rheobase_settings
        )

    efeatures, protocol_definitions, current = create_feature_protocol_files(
        cells,
        protocols,
        output_directory=output_directory,
        threshold_nvalue_save=threshold_nvalue_save,
        write_files=write_files,
        default_std_value=default_std_value
    )

    if pickle_cells:
        path_cells = pathlib.Path(output_directory)
        path_cells.mkdir(parents=True, exist_ok=True)
        pickle.dump(cells, open(path_cells / "cells.pkl", 'wb'))
        pickle.dump(protocols, open(path_cells / "protocols.pkl", 'wb'))

    if plot:
        plot_all_recordings_efeatures(
            cells, protocols, output_dir=output_directory, mapper=map_function
        )

    if extract_per_cell and write_files:
        extract_efeatures_per_cell(
            files_metadata,
            cells,
            output_directory,
            targets,
            protocol_mode,
            threshold_nvalue_save,
            write_files,
        )

    if not efeatures or not protocol_definitions:
        logger.warning("The output of the extraction is empty. Something went "
                       "wrong. Please check that your targets, files_metadata "
                       "and protocols_rheobase match the data you have "
                       "available.")

    return efeatures, protocol_definitions, current


def plot_recordings(
    files_metadata,
    output_directory="./figures/",
    recording_reader=None,
    map_function=map,
):
    """
    Plots recordings.

    Args:
        files_metadata (dict): define for which cell and protocol each file
            has to be used. Of the form:
            {
                cell_id: {
                    protocol_name: [{file_metadata1}, {file_metadata1}]
                }
            }
            A same file path might be present in the file metadata for
            different protocols.
            The entries required in the file_metadata are specific to each
            recording_reader (see bluepyemodel/reader.py to know which one
            are needed for your recording_reader).
        output_directory (str): path to the output directory where the plots
            will be saved.
        recording_reader (function): custom recording reader function. It's
            inner working has to match the metadata entered in files_metadata
        map_function (function): Function used to map (parallelize) the
            recording reading and feature extraction operations.
    """

    cells = read_recordings(
        files_metadata,
        recording_reader=recording_reader,
        map_function=map_function,
    )

    plot_all_recordings(cells, output_dir=output_directory)



"""Functions related to the computation of the rheobase"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
import logging
import numpy

logger = logging.getLogger(__name__)


def _get_list_spiking_amplitude(cell, protocols_rheobase):
    """Return the list of sorted list of amplitude that triggered at least
    one spike"""

    amps = []
    spike_counts = []

    for i, rec in enumerate(cell.recordings):
        if rec.protocol_name in protocols_rheobase:
            if rec.spikecount is not None:

                amps.append(rec.amp)
                spike_counts.append(rec.spikecount)

                if rec.amp < 0.01 and rec.spikecount >= 1:
                    logger.warning(
                        f"A recording of cell {cell.name} protocol "
                        f"{rec.protocol_name} shows spikes at a "
                        "suspiciously low current in a trace from file"
                        f" {rec.files}. Check that the ton and toff are"
                        "correct or for the presence of unwanted spikes."
                    )

    if amps:
        amps, spike_counts = zip(*sorted(zip(amps, spike_counts)))

    return amps, spike_counts


def compute_rheobase_absolute(cell, protocols_rheobase, spike_threshold=1):
    """ Compute the rheobase by finding the smallest current amplitude
    triggering at least one spike.

    Args:
        cell (Cell): cell for which to compute the rheobase
        protocols_rheobase (list): names of the protocols that will be
            used to compute the rheobase of the cells. E.g: ['IDthresh'].
        spike_threshold (int): number of spikes above which a recording
            is considered to compute the rheobase.
    """

    amps, spike_counts = _get_list_spiking_amplitude(cell, protocols_rheobase)

    for amp, spike_count in zip(amps, spike_counts):
        if spike_count >= spike_threshold:
            cell.rheobase = amp
            break


def compute_rheobase_flush(cell, protocols_rheobase, flush_length=1, upper_bound_spikecount=None):
    """ Compute the rheobase by finding the smallest current amplitude that:
        1. Triggered at least one spike
        2. Is followed by flush_length other traces that also trigger spikes.
    The advantage of this strategy is that it ignores traces showing spurious
    spikes at low amplitudes.
    Args:
        cell (Cell): cell for which to compute the rheobase
        protocols_rheobase (list): names of the protocols that will be
            used to compute the rheobase of the cells. E.g: ['IDthresh'].
        flush_length (int): number of traces that needs to show spikes for
            the candidate trace to be considered the rheobase.
        upper_bound_spikecount (int): if the spikecount of a recording is higher
            than this number, the recording will not trigger the start of a flush
    """

    amps, spike_counts = _get_list_spiking_amplitude(cell, protocols_rheobase)

    for i, amp in enumerate(amps):

        # We missed the threshold
        if upper_bound_spikecount is not None:
            if spike_counts[i] > upper_bound_spikecount:
                break

        if spike_counts[i]:

            end_flush = min(i + 1 + flush_length, len(amps))

            if (
                    numpy.count_nonzero(spike_counts[i + 1:end_flush]) == flush_length
            ):
                cell.rheobase = amp
                break


def compute_rheobase_majority_bin(
    cell, protocols_rheobase, min_step=0.01, majority=0.5
):
    """ Compute the rheobase by finding the smallest current amplitude
    triggering at least 1 spikes in the majority (default 50%) of the
    recordings.

    Args:
        cell (Cell): cell for which to compute the rheobase
        protocols_rheobase (list): names of the protocols that will be
            used to compute the rheobase of the cells. E.g: ['IDthresh'].
        min_step (float): minimum step above which amplitudes can be
            considered as separate steps
        majority (float): the proportion of sweeps with spike_threshold
            spikes to consider the target amplitude as rheobase
    """

    amps, spike_counts = _get_list_spiking_amplitude(cell, protocols_rheobase)

    bins = numpy.arange(min(amps), max(amps), min_step)
    bins_of_amps = numpy.digitize(amps, bins, right=False)

    for i, bin in enumerate(bins):

        spikes = [
            spike_counts[j] for j, idx in enumerate(bins_of_amps) if idx == i
        ]
        perc_spiking = numpy.mean([bool(s) for s in spikes])

        if perc_spiking >= majority:
            cell.rheobase = bin + (min_step / 2.)
            break


def compute_rheobase_interpolation(cell, protocols_rheobase):
    """ Compute the rheobase by fitting the reverse IF curve and finding the
    intersection with the line x = 1.

    Args:
        cell (Cell): cell for which to compute the rheobase
        protocols_rheobase (list): names of the protocols that will be
            used to compute the rheobase of the cells. E.g: ['IDthresh'].
    """

    amps, spike_counts = _get_list_spiking_amplitude(cell, protocols_rheobase)

    # Remove the excess zeros
    idx = next(
        (i for i in range(len(amps) - 1) if not spike_counts[i] and spike_counts[i + 1]),
        None
    )
    if idx is None:
        return
    amps = amps[idx:]
    spike_counts = spike_counts[idx:]

    if amps:

        try:
            fit = numpy.poly1d(numpy.polyfit(spike_counts, amps, deg=1))
        except:
            logger.error(
                f"Rheobase interpolation did not converge on cell {cell.name}"
            )
            return

        cell.rheobase = fit(1)



"""eCode init script"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project
 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>
 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.
 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.
 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from . import DeHyperPol
from . import HyperDePol
from . import SpikeRec
from . import negCheops
from . import posCheops
from . import ramp
from . import sAHP
from . import sineSpec
from . import step

# The ecode names have to be lower case only to avoid having to
# define duplicates for upper and lower cases.
eCodes = {
    "spontaneous": step.Step,
    "idrest": step.Step,
    "idthresh": step.Step,
    "idthres": step.Step,
    "apwaveform": step.Step,
    "iv": step.Step,
    "step": step.Step,
    "spontaps": step.Step,
    "firepattern": step.Step,
    "sponnohold30": step.Step,
    "sponhold30": step.Step,
    "starthold": step.Step,
    "startnohold": step.Step,
    "delta": step.Step,
    "sahp": sAHP.SAHP,
    "idhyperpol": sAHP.SAHP,
    "irdepol": sAHP.SAHP,
    "irhyperpol": sAHP.SAHP,
    "iddepol": sAHP.SAHP,
    "ramp": ramp.Ramp,
    "ap_thresh": ramp.Ramp,
    "apthresh": ramp.Ramp,
    "hyperdepol": HyperDePol.HyperDePol,
    "dehyperpol": DeHyperPol.DeHyperPol,
    "negcheops": negCheops.NegCheops,
    "poscheops": posCheops.PosCheops,
    "spikerec": SpikeRec.SpikeRec,
    "sinespec": sineSpec.SineSpec,
}



"""Tool and miscellaneous functions"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
import json
import numpy
import efel


DEFAULT_EFEL_SETTINGS = {
    'strict_stiminterval': True,
    'Threshold': -20.,
    'interp_step': 0.025
}


PRESET_PROTOCOLS_RHEOBASE = [
    "IV", "Step", "FirePattern", "IDrest", "IDRest", "IDthresh", "IDThresh", "IDThres", "IDthres"
]


def to_ms(t, t_unit):
    """Converts a time series to ms.

    Args:
        t (array): time series.
        t_unit (str): unit of the time series. Has to be "s", "sec",
            "seconds", "ms" or "10th_ms".
    """

    if t_unit.lower() in ["s", "sec", "seconds"]:
        return t * 1e3
    elif t_unit == "ms":
        return t
    elif t_unit == "10th_ms":
        return t * 0.1
    else:
        raise Exception("Time unit '{}' is unknown.".format(t_unit))


def to_nA(current, i_unit):
    """Converts a current series to nA.

    Args:
        current (array): current series.
        i_unit (str): unit of the current series. Has to be "a", "amperes",
            "amps", "mA", "uA", "pA" or "nA".
    """

    if i_unit.lower() in ["a", "amperes", "amps"]:
        return current * 1e9
    elif i_unit == "mA":
        return current * 1e6
    elif i_unit == "uA":
        return current * 1e3
    elif i_unit == "pA":
        return current * 1e-3
    elif i_unit == "nA":
        return current
    else:
        raise Exception("Current unit '{}' is unknown.".format(i_unit))


def to_mV(voltage, v_unit):
    """Converts a voltage series to mV.

    Args:
        voltage (array): voltage series.
        v_unit (str): unit of the voltage series. Has to be "v", "volts",
            "uV" or "mV".
    """

    if v_unit.lower() in ["v", "volts"]:
        return voltage * 1e3
    elif v_unit == "uV":
        return voltage * 1e-3
    elif v_unit == "mV":
        return voltage
    else:
        raise Exception("Voltage unit '{}' is unknown.".format(v_unit))


def set_efel_settings(efeature_settings):
    """Reset the eFEl settings and set them as requested by the user (uses
        default value otherwise).

    Args:
         efeature_settings (dict): eFEL settings in the form
            {setting_name: setting_value}.
    """

    efel.reset()

    for setting, value in efeature_settings.items():

        if setting in ['stim_start', 'stim_end']:
            value = float(value)

        if setting == 'Threshold':
            efel.setThreshold(value)

        elif isinstance(value, bool) or isinstance(value, int):
            efel.setIntSetting(setting, int(value))

        elif isinstance(value, (float, list)):
            efel.setDoubleSetting(setting, value)

        elif isinstance(value, str):
            efel.setStrSetting(setting, value)


def dict_to_json(data, path):
    """Save some data in a json file."""

    s = json.dumps(data, indent=2, cls=NumpyEncoder)
    with open(path, "w") as f:
        f.write(s)


class NumpyEncoder(json.JSONEncoder):

    """To make Numpy arrays JSON serializable"""

    def default(self, obj):
        if isinstance(
            obj,
            (
                numpy.int_,
                numpy.intc,
                numpy.intp,
                numpy.int8,
                numpy.int16,
                numpy.int32,
                numpy.int64,
                numpy.uint8,
                numpy.uint16,
                numpy.uint32,
                numpy.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(
            obj, (numpy.float_, numpy.float16, numpy.float32, numpy.float64)
        ):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



"""class AutoTarget"""

"""
Copyright (c) 2021, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


class AutoTarget():

    """Fuzzy targets defined by lists of ecodes, amplitudes and efeatures.
    The AutoTarget will try to find the best combination of ecodes, amplitudes
    and efeatures that match the data available in the recordings."""

    def __init__(
        self,
        protocols,
        amplitudes,
        efeatures,
        min_recordings_per_amplitude=10,
        preferred_number_protocols=1,
        tolerance=10,
    ):
        """
        Extract efeatures.

        Args:
            protocols (list of str): list of eCodes, by order of priority.
            amplitudes (list of int): list of amplitudes, expressed as
                percentages of the rheobase, by order of priority
            efeatures (list str): list of efeatures.
            min_recordings_per_amplitude (int): minimum number of recordings
                a amplitude should have to be considered as a target.
            preferred_number_protocols (int): out of all the available
                amplitudes, how many should be used.
            tolerance (float): tolerance used to bin the efeatures around the
                target amplitudes.
        """

        self.efeatures = efeatures
        self.protocols = protocols
        self.amplitudes = amplitudes

        self.min_recordings_per_amplitude = min_recordings_per_amplitude
        self.preferred_number_protocols = preferred_number_protocols
        self.tolerance = tolerance

        self.active_amplitudes = []
        self.active_ecode = []

    def select_ecode_and_amplitude(self, recordings):
        """Based on what ephys data is available, builds an effective (active)
         set of targets."""

        for ecode in self.protocols:

            available_amps = []
            for r in recordings:
                if r.protocol_name == ecode and r.amp_rel is not None:
                    available_amps.append(r.amp_rel)

            for amp in self.amplitudes:
                n_available_rec = sum(
                    abs(a - amp) < self.tolerance for a in available_amps
                )
                if n_available_rec >= self.min_recordings_per_amplitude:
                    self.active_ecode.append(ecode)
                    self.active_amplitudes.append(amp)

        if (
            self.active_ecode and
            len(self.active_ecode) > self.preferred_number_protocols
        ):
            self.active_ecode = self.active_ecode[
                :self.preferred_number_protocols]
            self.active_amplitudes = self.active_amplitudes[
                :self.preferred_number_protocols]

    def is_valid(self):
        """Check if the present AutoTarget has active targets (if matching
         ephys data were found)"""

        return bool(self.active_amplitudes) and bool(self.active_ecode)

    def generate_targets(self):
        """Build a list of targets in the format expected by the main
        extraction function of BluePyEfe using the targets presently
        active."""

        targets = []

        for amp, protocol_name in zip(
                self.active_amplitudes, self.active_ecode):
            for efeature in self.efeatures:
                targets.append({
                    "efeature": efeature,
                    "protocol": protocol_name,
                    "amplitude": amp,
                    "tolerance": self.tolerance,
                })

        return targets


def default_auto_targets():
    """Define a set of 3 generic AutoTarget for firing pattern properties,
     AP waveform properties and hyperpolarizing step properties."""

    auto_firing_pattern = AutoTarget(
        protocols=["Step", "FirePattern", "IDrest", "IDRest", "IDthresh",
                   "IDThresh", "IDThres", "IDthres", "IV"],
        amplitudes=[200, 150, 250, 300],
        efeatures=['voltage_base',
                   'adaptation_index2',
                   'mean_frequency',
                   'time_to_first_spike',
                   'time_to_last_spike',
                   'inv_first_ISI',
                   'inv_second_ISI',
                   'inv_third_ISI',
                   'inv_fourth_ISI',
                   'inv_fifth_ISI',
                   'inv_last_ISI',
                   'ISI_CV',
                   'ISI_log_slope',
                   'doublet_ISI',
                   'AP_amplitude',
                   'AP1_amp',
                   'APlast_amp',
                   'AHP_depth',
                   'AHP_time_from_peak'],
        min_recordings_per_amplitude=1,
        preferred_number_protocols=2,
        tolerance=25.,
    )

    auto_ap_waveform = AutoTarget(
        protocols=["APWaveform", "APwaveform", "Step", "FirePattern",
                   "IDrest", "IDRest", "IDthresh", "IDThresh", "IDThres",
                   "IDthres", "IV"],
        amplitudes=[300, 350, 250, 400, 200],
        efeatures=["AP_amplitude",
                   "AP1_amp",
                   "AP2_amp",
                   'AP_width',
                   "AP_duration_half_width",
                   "AP_rise_time",
                   "AP_fall_time",
                   'AHP_depth_abs',
                   'AHP_time_from_peak',
                   "AHP_depth"],
        min_recordings_per_amplitude=1,
        preferred_number_protocols=1,
        tolerance=25.,
    )

    auto_iv = AutoTarget(
        protocols=["IV", "Step"],
        amplitudes=[-50, -100],
        efeatures=['voltage_base',
                   'steady_state_voltage_stimend',
                   'ohmic_input_resistance_vb_ssse',
                   'voltage_deflection',
                   'voltage_deflection_begin',
                   'decay_time_constant_after_stim',
                   'Spikecount',
                   'sag_ratio1',
                   'sag_ratio2',
                   'sag_amplitude',
                   'sag_time_constant'],
        min_recordings_per_amplitude=1,
        preferred_number_protocols=1,
        tolerance=10.,
    )

    return [auto_firing_pattern, auto_ap_waveform, auto_iv]



"""sAHP eCode"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
import logging
import numpy

from ..recording import Recording
from .tools import scipy_signal2d
from .tools import base_current

logger = logging.getLogger(__name__)


class SAHP(Recording):

    """sAHP current stimulus

    The long step (here amp) is usually fixed at 40% of rheobase, and the short step (here amp2)
    can usually vary from 150% to 300% of rheobase.


       hypamp        hypamp+amp       hypamp+amp2        hypamp+amp           hypamp
         :                :                :                 :                   :
         :                :          ______________          :                   :
         :                :         |              |         :                   :
         :      ____________________|              |____________________         :
         :     |                    ^              ^                    |        :
         :     |                    :              :                    |        :
    |__________|                    :              :                    |__________________
    ^          ^                    :              :                    ^                  ^
    :          :                    :              :                    :                  :
    :          :                    :              :                    :                  :
    t=0        ton                  tmid           tmid2                toff               tend

    """

    def __init__(
        self,
        config_data,
        reader_data,
        protocol_name="sAHP",
        efel_settings=None
    ):

        super(SAHP, self).__init__(config_data, reader_data, protocol_name)

        self.ton = None
        self.tmid = None
        self.tmid2 = None
        self.toff = None
        self.tend = None
        self.amp = None
        self.amp2 = None
        self.hypamp = None
        self.dt = None

        self.amp_rel = None
        self.amp2_rel = None
        self.hypamp_rel = None

        if self.t is not None and self.current is not None:
            self.interpret(
                self.t, self.current, self.config_data, self.reader_data
            )

        if self.voltage is not None:
            self.set_autothreshold()
            self.compute_spikecount(efel_settings)

        self.export_attr = ["ton", "tmid", "tmid2", "toff", "tend", "amp",
                            "amp2", "hypamp", "dt", "amp_rel", "amp2_rel",
                            "hypamp_rel"]

    def get_stimulus_parameters(self):
        """Returns the eCode parameters"""
        ecode_params = {
            "delay": self.ton,
            "tmid": self.tmid,
            "tmid2": self.tmid2,
            "toff": self.toff,
            "amp": self.amp2,
            "long_amp": self.amp,
            "thresh_perc": self.amp2_rel,
            "duration": self.tmid2 - self.tmid,
            "totduration": self.tend,
        }
        return ecode_params

    def compute_amp(self, current, config_data, reader_data):

        smooth_current = scipy_signal2d(current, 85)

        hypamp_value = numpy.median(
            numpy.concatenate(
                (smooth_current[: self.ton], smooth_current[self.toff :])
            )
        )
        self.set_amplitudes_ecode("hypamp", config_data, reader_data, hypamp_value)

        amp_value = numpy.median(
            numpy.concatenate(
                (
                    smooth_current[self.ton: self.tmid],
                    smooth_current[self.tmid2: self.toff],
                )
            )
        ) - self.hypamp
        self.set_amplitudes_ecode("amp", config_data, reader_data, amp_value)

        amp2_value = numpy.median(smooth_current[self.tmid : self.tmid2]) - self.hypamp
        self.set_amplitudes_ecode("amp2", config_data, reader_data, amp2_value)

        if config_data.get("tend", None) is None:
            self.tend = len(self.t) * self.dt
        else:
            self.tend = config_data["tend"]

        self.ton = self.t[int(round(self.ton))]
        self.toff = self.t[int(round(self.toff))]
        self.tmid = self.t[int(round(self.tmid))]
        self.tmid2 = self.t[int(round(self.tmid2))]

    def step_detection(self, current, config_data, reader_data):

        # Set the threshold to detect the step
        noise_level = numpy.std(numpy.concatenate((self.current[:50], self.current[-50:])))
        step_threshold = numpy.max([2.0 * noise_level, 1e-5])

        # The buffer prevent miss-detection of the step when artifacts are
        # present at the very start or very end of the current trace
        buffer_detect = 2.0
        idx_buffer = int(buffer_detect / self.dt)
        idx_buffer = max(1, idx_buffer)

        buffer_step = 50
        smooth_current = scipy_signal2d(current, 85)

        # Infer the beginning of the long step
        self.hypamp = base_current(current)

        if "ton" in config_data and config_data["ton"] is not None:
            self.ton = int(round(config_data["ton"] / self.dt))
        else:
            tmp_current = numpy.abs(smooth_current[idx_buffer:] - self.hypamp)
            self.ton = idx_buffer + numpy.argmax(tmp_current > step_threshold)

        # Infer the end of the long step
        tmp_current = numpy.flip(
            numpy.abs(smooth_current[self.ton:-idx_buffer] - self.hypamp)
        )
        self.toff = (
            (len(current) - numpy.argmax(tmp_current > step_threshold)) - 1 - idx_buffer
        )

        # Get the amplitude of the step current (relative to hypamp)
        self.amp = numpy.median(
            numpy.concatenate((smooth_current[self.ton:self.ton + 50],
                               smooth_current[self.toff - 50:self.toff]))
        ) - self.hypamp

        # Infer the beginning of the short step
        tmp_current = numpy.abs(
            smooth_current[self.ton + buffer_step:self.toff - buffer_step] -
            self.amp - self.hypamp
        )
        self.tmid = self.ton + buffer_step + numpy.argmax(tmp_current > step_threshold)

        # Infer the end of the long step
        tmp_current = numpy.flip(
            numpy.abs(
                smooth_current[self.ton + buffer_step:self.toff - 50] -
                self.amp - self.hypamp
            )
        )
        self.tmid2 = (
            (self.toff - numpy.argmax(tmp_current > step_threshold)) - 1 - buffer_step
        )

        self.amp2 = numpy.median(smooth_current[self.tmid:self.tmid2]) - self.hypamp

        # Converting back ton and toff to ms
        self.ton = self.t[int(round(self.ton))]
        self.toff = self.t[int(round(self.toff))]
        self.tmid = self.t[int(round(self.tmid))]
        self.tmid2 = self.t[int(round(self.tmid2))]
        self.tend = len(self.t) * self.dt

        # Check for some common step detection failures when the current
        # is constant.
        if self.ton >= self.toff or self.ton >= self.tend or \
                self.toff > self.tend or self.tmid == self.ton \
                or self.tmid2 == self.toff:

            self.ton = 0.
            self.toff = self.tend

            logger.warning(
                "The automatic step detection failed for the recording "
                f"{self.protocol_name} in files {self.files}. You should "
                "specify ton and toff by hand in your files_metadata "
                "for this file."
            )

    def interpret(self, t, current, config_data, reader_data):
        """Analyse a current with a step and extract from it the parameters
        needed to reconstruct the array"""

        self.dt = t[1]
        required = ["ton", "tmid", "tmid2", "toff"]
        if all(r in config_data for r in required):
            self.set_timing_ecode(required, config_data)
            self.compute_amp(current, config_data, reader_data)
        else:
            self.step_detection(current, config_data, reader_data)

    def generate(self):
        """Generate the current array from the parameters of the ecode"""

        ton = int(self.ton / self.dt)
        tmid = int(self.tmid / self.dt)
        tmid2 = int(self.tmid2 / self.dt)
        toff = int(self.toff / self.dt)

        time = numpy.arange(0.0, self.tend, self.dt)
        current = numpy.full(time.shape, self.hypamp)
        current[ton:tmid] += self.amp
        current[tmid2:toff] += self.amp
        current[tmid:tmid2] += self.amp2

        return time, current

    def compute_relative_amp(self, amp_threshold):
        self.amp_rel = 100.0 * self.amp / amp_threshold
        self.amp2_rel = 100.0 * self.amp2 / amp_threshold
        self.hypamp_rel = 100.0 * self.hypamp / amp_threshold

    def in_target(self, target, tolerance, absolute_amplitude=False):
        """Returns a boolean. True if the amplitude of the eCode is close to
        target and False otherwise."""

        effective_amp = self.amp2 if absolute_amplitude else self.amp2_rel

        if numpy.abs(target - effective_amp) < tolerance:
            return True

        return False

    def get_plot_amplitude_title(self):
        return " ({:.01f}%/{:.01f}%)".format(self.amp_rel, self.amp2_rel)



"""Protocol class"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy
import logging

from bluepyefe.ecode import eCodes

logger = logging.getLogger(__name__)


class Protocol():

    """Protocol informs about the current stimulus that was used to obtain
     efeatures at a given amplitude for a given protocol name. This class
     is mainly used to produce a description of the experimental protocol
     that can be used in BluePyOpt"""

    def __init__(
        self,
        name,
        amplitude,
        tolerance,
        feature_targets=None,
        global_rheobase=None,
        mode="mean",
    ):
        """Constructor

        Args:
            name (str): name of the protocol (ex: 'APWaveform')
            amplitude (float): amplitude of the current stimuli for the
                present protocol (expressed as a percentage of the
                threshold amplitude or in absolute current depending on the
                setting absolute_amplitude)
            tolerance (float): tolerance around the target amplitude in which
                an experimental recording will be seen as a hit during
                efeatures extraction (expressed as a percentage of the
                threshold amplitude or in absolute current depending on the
                setting absolute_amplitude)
            feature_targets (list): list of EFeatureTarget associated to the
                protocol
            global_rheobase (float): average rheobase across all cells
            mode (str): if the protocol matches several recordings, the mode
                set the logic of how the output will be generating. Must be
                'mean', 'median' or 'lnmc'
        """

        self.name = name
        self.amplitude = amplitude
        self.tolerance = tolerance

        self.feature_targets = feature_targets
        if self.feature_targets is None:
            self.feature_targets = []

        self.global_rheobase = global_rheobase
        self.mode = mode

        self.recordings = []

    @property
    def stimulus_name(self):
        """Name of the stimulus associated to the protocol"""

        return f"{self.name}_{self.amplitude}"

    @property
    def n_match(self):
        """Number of recordings whose amplitude matched the present protocol"""

        return sum([f.sample_size for f in self.feature_targets])

    @property
    def ecode(self):
        """Create a temporary eCode that matches all the recordings for the
        present protocol. The eCode's parameters are computed differently
        depending on the mode of the protocol"""

        if not self.recordings:
            return None

        for ecode in eCodes.keys():
            if ecode.lower() in self.name.lower():
                ecode = eCodes[ecode]({}, {}, self.name)
                break
        else:
            raise KeyError(
                "There is no eCode linked to the stimulus name {}. See "
                "ecode/__init__.py for the available stimuli names"
                "".format(self.name.lower())
            )

        if self.mode == "mean":
            self.reduce_ecode(ecode, operator=numpy.nanmean)
        elif self.mode == "median":
            self.reduce_ecode(ecode, operator=numpy.nanmedian)
        elif self.mode == "min":
            self.reduce_ecode(ecode, operator=numpy.nanmin)
        elif self.mode == "max":
            self.reduce_ecode(ecode, operator=numpy.nanmax)
        else:
            raise ValueError("'mode' should be mean or median")

        return ecode

    def append(self, recording):
        """Append a Recording to the present protocol"""

        for i, target in enumerate(self.feature_targets):
            if target.efeature_name in recording.efeatures:
                self.feature_targets[i].append(
                    recording.efeatures[target.efeature_name],
                    recording.files
                )

            if (
                recording.auto_threshold is not None and
                "Threshold" not in self.feature_targets[i].efel_settings
            ):
                self.feature_targets[i]._auto_thresholds.append(
                    recording.auto_threshold)

        self.recordings.append(recording)

    def as_dict(self):
        """Returns a dictionary that defines the present protocol. This
        definition is computed differently depending on the mode of the
        protocol
        """

        return {
            "holding": {
                "delay": 0.0,
                "amp": self.ecode.hypamp,
                "duration": self.ecode.tend,
                "totduration": self.ecode.tend,
            },
            "step": self.ecode.get_stimulus_parameters(),
        }

    def reduce_ecode(self, ecode, operator):
        """Creates an eCode defined from the parameters of all the recordings
        matching the present protocol"""

        if not self.recordings:
            logger.warning(
                "Could not compute average ecode for protocol {} target {} "
                "because it didn't match any recordings".format(
                    self.name, self.amplitude
                )
            )
            return None

        params = [r.get_params() for r in self.recordings]

        if self.global_rheobase is None and "amp" in params[0]:
            logger.warning(
                "No global threshold amplitude passed. This can result"
                " in inconsistencies in-between protocols if some cells "
                "only matched a subset of the targets."
            )

        for key in params[0]:

            if isinstance(params[0][key], (list, numpy.ndarray)):
                logger.warning(
                    "Parameter {} for protocol {} is a list and cannot be "
                    "averaged across recordings".format(key, self.name)
                )
                setattr(ecode, key, params[0][key])
                continue

            if key == "amp" and self.global_rheobase:
                amp_rel = operator([c["amp_rel"] for c in params])
                mean_param = float(amp_rel) * self.global_rheobase / 100.
            elif key == "amp2" and self.global_rheobase:
                amp_rel2 = operator([c["amp2_rel"] for c in params])
                mean_param = float(amp_rel2) * self.global_rheobase / 100.
            else:
                mean_param = operator([numpy.nan if c[key] is None else c[key] for c in params])

            setattr(ecode, key, mean_param)

        return ecode

    def __str__(self):
        """String representation"""

        str_form = "Protocol {} {:.1f}%:\n".format(
            self.name, self.amplitude
        )

        str_form += "Number of matching recordings: {}".format(self.n_match)

        if self.n_match:
            str_form += "\neCode: {}\n".format(self.as_dict)

        return str_form



"""Plotting functions"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import logging
import pathlib
from itertools import cycle
from functools import partial

import matplotlib.pyplot as plt
import numpy

logger = logging.getLogger(__name__)


def _save_fig(directory, filename):
    """Save a matplotlib figure.

    Args:
        directory (str): path of the directory in which to save the figure.
        filename (str): name of the file in which to save the figure.
    """

    dir_path = pathlib.Path(directory)
    dir_path.mkdir(parents=True, exist_ok=True)

    fig_path = dir_path / filename

    plt.savefig(fig_path, dpi=80, pad_inches=0)
    plt.close("all")
    plt.clf()


def _get_colors_markers_wheels(cells):
    """Generate a color and a marker dictionary unique to each cell.

    Args:
        cells (list of Cells): list of the cells
    """

    colors = cycle(["C{}".format(i) for i in range(10)])
    markers = cycle(["o", "v", "^", "s", "*", "d", "x"])

    cell_names = [c.name for c in cells]

    colors = {n: next(colors) for n in cell_names}
    markers = {n: next(markers) for n in cell_names}

    return colors, markers


def _plot_legend(colors, markers, output_dir, show=False):
    """Draw a separate legend figure"""

    if not len(colors):
        raise Exception("Plot legend needs an non-empty colors dictionary.")
    if not len(markers):
        raise Exception("Plot legend needs an non-empty markers dictionary.")

    ncols = numpy.ceil(numpy.sqrt(len(colors)))

    fig, axs = plt.subplots(1, figsize=(6, 6), squeeze=False)

    for i, (cellname, c, m) in enumerate(
        zip(colors.keys(), colors.values(), markers.values())
    ):
        axs[0, 0].scatter(
            x=2.5 * int(i / ncols), y=0.5 * (i % ncols), c=c, marker=m, s=8.0
        )

        axs[0, 0].text(
            x=2.5 * int(i / ncols) + 0.1,
            y=0.5 * (i % ncols) - 0.08,
            s=cellname,
            fontsize=5,
        )

    axs[0, 0].set_xlim(-1.0, 2.0 * ncols + 1)
    axs[0, 0].set_ylim(-1.0, ncols + 1)

    axs[0, 0].set_xticks([])
    axs[0, 0].set_yticks([])

    axs[0, 0].set_frame_on(False)

    if show:
        fig.show()

    if output_dir:
        _save_fig(output_dir, "legend.pdf")

    return fig, axs


def _plot(cell, output_dir, show=False):
    cell.plot_all_recordings(output_dir, show=show)


def plot_all_recordings(cells, output_dir, show=False, mapper=map):
    """Plot recordings for all cells and all protocols"""
    if mapper == map:
        # For the built-in map(), ensure immediate evaluation as it returns a lazy iterator
        # which won't execute the function until iterated over. Converting to a list forces this iteration.
        list(mapper(partial(_plot, output_dir=output_dir, show=show), cells))
    else:
        mapper(partial(_plot, output_dir=output_dir, show=show), cells)


def plot_efeature(
    cells,
    efeature,
    protocol_name,
    output_dir,
    protocols=[],
    key_amp="amp",
    colors=None,
    markers=None,
    show=False,
    show_targets=True,
):
    """Plot one efeature for a protocol"""

    if not cells:
        logger.warning("In plot_efeature, no cells object to plot.")
        return None

    if not colors or not markers:
        colors, markers = _get_colors_markers_wheels(cells)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    has_data = False

    for cell in cells:

        x, y = [], []
        for rec in cell.get_recordings_by_protocol_name(protocol_name):
            if hasattr(rec, key_amp) and getattr(rec, key_amp):
                if efeature in rec.efeatures:
                    x.append(getattr(rec, key_amp))
                    y.append(rec.efeatures[efeature])

        if y:
            has_data = True

        ax.scatter(
            x, y, c=colors[cell.name], marker=markers[cell.name], s=5.0
        )

        for protocol in protocols:
            if protocol.name == protocol_name:

                if key_amp == "amp_rel":
                    x_key = protocol.amplitude
                elif key_amp == "amp":
                    x_key = numpy.mean([t.amp for t in protocol.recordings])

                target = next(
                    (t for t in protocol.feature_targets if
                     t.efel_feature_name == efeature),
                    None
                )
                if target is not None and show_targets:
                    ax.errorbar(
                        x_key,
                        target.mean,
                        yerr=target.std,
                        marker="o",
                        elinewidth=0.8,
                        markersize=3.0,
                        c="black",
                        zorder=100,
                        alpha=0.8,
                        capsize=2
                    )

    if not has_data:
        return

    if key_amp == "amp_rel":
        ax.set_xlabel(
            "Relative step amplitude" " ($I/I_{thresh}$)", size="x-large"
        )
    elif key_amp == "amp":
        ax.set_xlabel("Step amplitude (nA)", size="x-large")
    ax.set_ylabel(efeature, size="x-large")
    ax.set_title(
        f"Protocol: {protocol_name}, EFeature: {efeature}", size="xx-large"
    )

    if show:
        fig.show()

    if output_dir:
        if len(cells) == 1:
            filename = "{}_{}_{}_{}.pdf".format(
                cells[0].name, protocol_name, efeature, key_amp
            )
            dirname = pathlib.Path(output_dir) / cells[0].name
        else:
            filename = "{}_{}_{}.pdf".format(protocol_name, efeature, key_amp)
            dirname = pathlib.Path(output_dir)

    _save_fig(dirname, filename)


def plot_efeatures(
    cells,
    protocol_name,
    output_dir=None,
    protocols=[],
    key_amp="amp",
    colors=None,
    markers=None,
    show=False,
    show_targets=True,
):
    """
    Plot the efeatures of a cell or a group of cells versus current amplitude
    or relative current amplitude.
    """

    if not cells:
        logger.warning("In plot_efeatures, no cells object to plot.")
        return None, None

    efeatures = set()
    for p in protocols:
        if p.name == protocol_name:
            efeatures.update([t.efel_feature_name for t in p.feature_targets])

    for fi, efeature in enumerate(efeatures):
        plot_efeature(
            cells,
            efeature,
            protocol_name,
            output_dir,
            protocols=protocols,
            key_amp=key_amp,
            colors=colors,
            markers=markers,
            show=show,
            show_targets=show_targets,
        )


def _plot_ind(cell, output_dir, protocols, key_amp, colors, markers, show):
    for protocol_name in cell.get_protocol_names():
        _ = plot_efeatures(
            cells=[cell],
            protocol_name=protocol_name,
            output_dir=output_dir,
            protocols=protocols,
            key_amp=key_amp,
            colors=colors,
            markers=markers,
            show=show,
            show_targets=False
        )


def plot_individual_efeatures(
    cells,
    protocols,
    output_dir=None,
    colors=None,
    markers=None,
    key_amp="amp",
    show=False,
    mapper=map,
):
    """Generate efeatures plots for all each cell individually"""

    if not colors or not markers:
        colors, markers = _get_colors_markers_wheels(cells)

    mapper(
        partial(
            _plot_ind,
            output_dir=output_dir,
            protocols=protocols,
            key_amp=key_amp,
            colors=colors,
            markers=markers,
            show=show,
        ),
        cells,
    )


def plot_grouped_efeatures(
    cells,
    protocols,
    output_dir=None,
    colors=None,
    markers=None,
    key_amp="amp",
    show=False
):
    """Generate plots for each efeature across all cells."""

    if not cells:
        logger.warning("In plot_grouped_efeatures, no cells object to plot.")
        return

    if not colors or not markers:
        colors, markers = _get_colors_markers_wheels(cells)

    protocol_names = set([p.name for p in protocols])

    for protocol_name in protocol_names:

        _ = plot_efeatures(
            cells=cells,
            protocol_name=protocol_name,
            output_dir=output_dir,
            protocols=protocols,
            key_amp=key_amp,
            colors=colors,
            markers=markers,
            show=show
        )

    _ = _plot_legend(colors, markers, output_dir, show=show)


def plot_all_recordings_efeatures(
    cells, protocols, output_dir=None, show=False, mapper=map
):
    """Generate plots for all recordings and efeatures both for individual
    cells and across all cells."""

    colors, markers = _get_colors_markers_wheels(cells)

    plot_all_recordings(cells, output_dir, mapper=mapper)

    for key_amp in ["amp", "amp_rel"]:
        plot_individual_efeatures(
            cells,
            protocols,
            output_dir,
            colors=colors,
            markers=markers,
            key_amp=key_amp,
            show=show,
            mapper=mapper,
        )
        plot_grouped_efeatures(
            cells,
            protocols,
            output_dir,
            colors=colors,
            markers=markers,
            key_amp=key_amp,
            show=show
        )



"""To translate config dictionaries from BluePyEfe 1 to input needed by BluePyEfe 2"""

"""
Copyright (c) 2022, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import pathlib


def translate_legacy_files_metadata(config):
    """Translate the legacy field "cells" into the new files_metadata"""

    if "path" in config:
        ephys_path = pathlib.Path(config["path"])
    else:
        ephys_path = pathlib.Path("./")

    files_metadata = {}
    for cell_name in config["cells"]:
        for protocol_name in config["cells"][cell_name]["experiments"]:
            for file_metadata in config["cells"][cell_name]["experiments"][protocol_name]["files"]:

                if cell_name not in files_metadata:
                    files_metadata[cell_name] = {}
                if protocol_name not in files_metadata[cell_name]:
                    files_metadata[cell_name][protocol_name] = []

                if "i_file" in file_metadata:
                    filepaths = {
                        "i_file": str(ephys_path / file_metadata["i_file"]),
                        "v_file": str(ephys_path / file_metadata["v_file"]),
                    }
                else:
                    filepaths = {"filepath": str(ephys_path / file_metadata["filepath"])}

                files_metadata[cell_name][protocol_name].append(file_metadata)
                files_metadata[cell_name][protocol_name][-1].update(filepaths)

                if protocol_name in config["options"]["onoff"]:
                    files_metadata[cell_name][protocol_name][-1]["ton"] = config[
                        "options"]["onoff"][protocol_name][0]
                    files_metadata[cell_name][protocol_name][-1]["toff"] = config[
                        "options"]["onoff"][protocol_name][1]

    return files_metadata


def translate_legacy_targets(config):
    """Translate the legacy field "targets" into the new targets"""

    targets = []

    for protocol in config["features"]:
        for feature in config["features"][protocol]:
            if "spikerate" in feature:
                continue
            for amp, tol in zip(config["options"]["target"], config["options"]["tolerance"]):

                if amp == "all":
                    continue
                if amp == "noinput":
                    effective_amp = 0
                    effective_tolerance = 10
                else:
                    effective_amp = amp
                    effective_tolerance = tol

                efel_settings = {}
                if "strict_stiminterval" in config["options"]:
                    if protocol in config["options"]["strict_stiminterval"]:
                        efel_settings["strict_stiminterval"] = config["options"][
                            "strict_stiminterval"][protocol]
                    elif "base" in config["options"]["strict_stiminterval"]:
                        efel_settings["strict_stiminterval"] = config["options"][
                            "strict_stiminterval"]["base"]

                targets.append(
                    {
                        "efeature": feature,
                        "protocol": protocol,
                        "amplitude": effective_amp,
                        "tolerance": effective_tolerance,
                        "efel_settings": efel_settings
                    }
                )

    return targets


def translate_legacy_config(config):
    """Translate a legacy config from BluePyEfe 1 to BluePyEfe 2"""

    files_metadata = translate_legacy_files_metadata(config)
    targets = translate_legacy_targets(config)
    protocols_rheobase = config["options"]["expthreshold"]

    rheobase_strategy = "absolute"
    rheobase_settings = {}
    if "spike_threshold" in config["options"]:
        rheobase_settings["spike_threshold"] = config["options"]["spike_threshold"]

    return {
        "files_metadata": files_metadata,
        "targets": targets,
        "protocols_rheobase": protocols_rheobase,
        "rheobase_strategy": rheobase_strategy,
        "rheobase_settings": rheobase_settings
    }



import logging
import numpy

logger = logging.getLogger(__name__)


class NWBReader:
    def __init__(self, content, target_protocols, repetition=None):
        """ Init

        Args:
            content (h5.File): NWB file
            target_protocols (list of str): list of the protocols to be read and returned
            repetition (list of int): id of the repetition(s) to be read and returned
        """

        self.content = content
        self.target_protocols = target_protocols
        self.repetition = repetition

    def read(self):
        """ Read the content of the NWB file

        Returns:
            data (list of dict): list of traces"""

        raise NotImplementedError()

    def _format_nwb_trace(self, voltage, current, start_time, trace_name=None, repetition=None):
        """ Format the data from the NWB file to the format used by BluePyEfe

        Args:
            voltage (Dataset): voltage series
            current (Dataset): current series
            start_time (Dataset): starting time
            trace_name (Dataset): name of the trace
            repetition (int): repetition number

        Returns:
            dict: formatted trace
        """

        v_array = numpy.array(
            voltage[()] * voltage.attrs["conversion"], dtype="float32"
        )

        i_array = numpy.array(
            current[()] * current.attrs["conversion"], dtype="float32"
        )

        dt = 1. / float(start_time.attrs["rate"])

        v_unit = voltage.attrs["unit"]
        i_unit = current.attrs["unit"]
        t_unit = start_time.attrs["unit"]
        if not isinstance(v_unit, str):
            v_unit = voltage.attrs["unit"].decode('UTF-8')
            i_unit = current.attrs["unit"].decode('UTF-8')
            t_unit = start_time.attrs["unit"].decode('UTF-8')

        return {
            "voltage": v_array,
            "current": i_array,
            "dt": dt,
            "id": str(trace_name),
            "repetition": repetition,
            "i_unit": i_unit,
            "v_unit": v_unit,
            "t_unit": t_unit,
        }


class AIBSNWBReader(NWBReader):
    def read(self):
        """ Read the content of the NWB file

        Returns:
            data (list of dict): list of traces"""

        data = []

        for sweep in list(self.content["acquisition"]["timeseries"].keys()):
            protocol_name = self.content["acquisition"]["timeseries"][sweep]["aibs_stimulus_name"][()]
            if not isinstance(protocol_name, str):
                protocol_name = protocol_name.decode('UTF-8')

            if self.target_protocols and protocol_name not in self.target_protocols:
                continue

            data.append(self._format_nwb_trace(
                voltage=self.content["acquisition"]["timeseries"][sweep]["data"],
                current=self.content["stimulus"]["presentation"][sweep]["data"],
                start_time=self.content["acquisition"]["timeseries"][sweep]["starting_time"],
                trace_name=sweep
            ))

        return data


class ScalaNWBReader(NWBReader):
    def read(self):
        """ Read and format the content of the NWB file

        Returns:
            data (list of dict): list of traces
        """

        data = []

        for sweep in list(self.content['acquisition'].keys()):
            key_current = sweep.replace('Series', 'StimulusSeries')
            protocol_name = "Step"

            if self.target_protocols and protocol_name not in self.target_protocols:
                continue

            if key_current not in self.content['stimulus']['presentation']:
                continue

            data.append(self._format_nwb_trace(
                voltage=self.content['acquisition'][sweep]['data'],
                current=self.content['stimulus']['presentation'][key_current]['data'],
                start_time=self.content["acquisition"][sweep]["starting_time"],
                trace_name=sweep,
            ))

        return data


class BBPNWBReader(NWBReader):
    def _get_repetition_keys_nwb(self, ecode_content, request_repetitions=None):
        """ Filter the names of the traces based on the requested repetitions

        Args:
            ecode_content (dict): content of the NWB file for one eCode/protocol
            request_repetitions (list of int): identifier of the requested repetitions

        Returns:
            list of str: list of the keys of the traces to be read
        """

        if isinstance(request_repetitions, (int, str)):
            request_repetitions = [int(request_repetitions)]

        reps = list(ecode_content.keys())
        reps_id = [int(rep.replace("repetition ", "")) for rep in reps]

        if request_repetitions:
            return [reps[reps_id.index(i)] for i in request_repetitions]
        else:
            return list(ecode_content.keys())

    def read(self):
        """ Read and format the content of the NWB file

        Returns:
            data (list of dict): list of traces
        """

        data = []

        for ecode in self.target_protocols:
            for cell_id in self.content["data_organization"].keys():
                if ecode not in self.content["data_organization"][cell_id]:
                    logger.debug(f"No eCode {ecode} in nwb.")
                    continue

                ecode_content = self.content["data_organization"][cell_id][ecode]

                rep_iter = self._get_repetition_keys_nwb(
                    ecode_content, request_repetitions=self.repetition
                )

                for rep in rep_iter:
                    for sweep in ecode_content[rep].keys():
                        for trace_name in list(ecode_content[rep][sweep].keys()):
                            if "ccs_" in trace_name:
                                key_current = trace_name.replace("ccs_", "ccss_")
                            elif "ic_" in trace_name:
                                key_current = trace_name.replace("ic_", "ics_")
                            else:
                                continue

                            if key_current not in self.content["stimulus"]["presentation"]:
                                logger.debug(f"Ignoring {key_current} not"
                                             " present in the stimulus presentation")
                                continue
                            if trace_name not in self.content["acquisition"]:
                                logger.debug(f"Ignoring {trace_name} not"
                                             " present in the acquisition")
                                continue

                            data.append(self._format_nwb_trace(
                                voltage=self.content["acquisition"][trace_name]["data"],
                                current=self.content["stimulus"]["presentation"][key_current][
                                    "data"],
                                start_time=self.content["stimulus"]["presentation"][key_current][
                                    "starting_time"],
                                trace_name=trace_name,
                                repetition=int(rep.replace("repetition ", ""))
                            ))

        return data
